module iohev

    interface
        subroutine readdataset2 ( uid, datasetf, inputf, mapf, frqtqf, dt, nidle, nrate )
            character (len=*) inputf, datasetf, mapf, frqtqf
            real(8) dt, nidle, nrate
            integer uid
        end subroutine

        subroutine readrunningdata ( uid, fname, ndata, t,v,ne,tq, sp )
            integer uid, ndata
            character (len=*) fname                                       ! torque filename
            real(8), pointer :: t(:), v(:), ne(:), tq(:)
            integer, pointer :: sp(:)
        end subroutine

        subroutine readtqfc ( uid, tqf, tqc )
            use hvtransfc
            integer uid
            type (stqcurve) tqc
            character (len=*) tqf
        end subroutine readtqfc

        subroutine readmapfc( uid, mapf, fcm )
            use hvtransfc
            integer uid
            type (sfcmap) fcm
            character (len=*) mapf
        end subroutine readmapfc

        subroutine outaveragefc ( uid, fname, inputf, mapf, ftqf, dist, fuelc, avgv, avgfc )
            integer uid
            character (len=*) fname, inputf, mapf, ftqf
            real(8) avgv, avgfc, dist, fuelc
        end subroutine

        subroutine outtimeseriesfc (uid,fname,ndata,t, vreal,ne,te,sp,fc)
            integer uid, sp(:), ndata
            character (len=*) fname
            real(8) t(:), vreal(:), ne(:), te(:), fc(:)
        end subroutine

    end interface

end module iohev



!**************************************************************
! SUBROUTINE READSPECFC
!  Read vehicle specification data
!**************************************************************
subroutine readdataset2 ( uid, datasetf, inputf, mapf, frqtqf,dt, nidle, nrate )
    implicit none

    character (len=*) inputf, datasetf, mapf, frqtqf
    integer uid, ios
    real(8) dt, nidle, nrate

    open ( uid, file=trim(datasetf), status='OLD', access='sequential', iostat=ios )
    if ( ios /= 0 ) then
        write(0,*) ' Cannot open dataset file : ', trim(datasetf)
        stop
    end if

    read ( uid, * ) inputf                                      ! vehicle name
    read ( uid, * ) dt                                          ! ԍ
    read ( uid, * ) mapf                                        ! type name
    read ( uid, * ) frqtqf                                      ! engine filename
    read ( uid, * ) nidle                                       ! AChO] 0
    read ( uid, * ) nrate                                       ! i] 100

    close(uid)

end subroutine


! *************************************************************
! SUBROUTINE READTQFC
!  Read engine torque input data
!**************************************************************
subroutine readrunningdata ( uid, fname, ndata, t, v, ne, tq, sp )
    implicit none

    integer uid, ndata
    character (len=*) fname                                         ! sf[^t@C
    real(8), pointer :: t(:), v(:), ne(:), tq(:)
    integer, pointer :: sp(:)

    character (len=1024) tmp
    integer i, ios, spt
    real(8) tt, vt, net, tqt

!   -----------------------------------------------------------
!   count number of data
!   -----------------------------------------------------------
    ndata = 0
    open ( uid, file = fname, status = 'old', err=100 )
    read ( uid,'(A)' ) tmp                                              ! wb_

    ios = 0
    do while ( ios == 0 )
        read ( uid, *, iostat = ios, err=110 ) tt, vt, net, tqt, spt    ! ꎞǂݍ
        if( ios == 0 ) ndata = ndata + 1                                ! f[^JEg
    end do

!   -----------------------------------------------------------
!   read data
!   -----------------------------------------------------------
    allocate ( t( ndata ) )
    allocate ( v( ndata ) )
    allocate ( ne( ndata ) )
    allocate ( tq( ndata ) )
    allocate ( sp( ndata ) )

    rewind ( uid )                                                      ! 擪ֈړ
    read ( uid,'(A)' ) tmp                                              ! wb_

    do i = 1, ndata
        read ( uid, *, err=110 ) t(i), v(i) , ne(i), tq(i), sp(i)       ! sf[^ǂݍ
    end do

    close ( uid )

    return

!   -----------------------------------------------------------
!   G[Ή
!   -----------------------------------------------------------
100 write (0,'(3A)') " Cannot open running data file [ ", trim(fname), " ]"
    stop
110 write (0,'(A,i0,3A)') " Failed to read running data. Data No.= ",    &
                          ndata+1, " in [ ", trim( fname ), " ]"
    close ( uid )
    stop

end subroutine





!**************************************************************
! SUBROUTINE READTQFC
!  Read engine torque input data
!**************************************************************
subroutine readtqfc ( uid, tqf, tqc )
    use hvtransfc
    implicit none

    integer uid
    character (len=*) tqf                                       ! torque filename
    type (stqcurve) tqc                                         ! torque curve

    character (len=1024) tmp
    integer i, j, gap, ios
    real(8) revt, tmaxt

!   -----------------------------------------------------------
!   count number of torque data
!   -----------------------------------------------------------
    tqc%ndata = 0                                               ! initialize
    open ( uid, file = tqf, status = 'old', err=600 )
    read ( uid,'(A)' ) tmp                                      ! skip header
    read ( uid,'(A)' ) tmp                                      ! skip header

    ios = 0                                                     ! status
    do while ( ios == 0 )
        read ( uid, *, iostat = ios, err=650 ) revt, tmaxt      ! test reading of data set
        if( ios == 0 ) tqc%ndata = tqc%ndata + 1                ! count n of data
    end do

!   -----------------------------------------------------------
!   read torque data
!   -----------------------------------------------------------
    allocate ( tqc%tq ( tqc%ndata ) )
    allocate ( tqc%rev( tqc%ndata ) )

    rewind ( uid )                                              ! move to head of file
    read ( uid,'(A)' ) tmp                                      ! skip header
    read ( uid,'(A)' ) tmp                                      ! skip header

    do i = 1, tqc%ndata
        read ( uid, *, err=650 ) tqc%rev(i), tqc%tq(i)          ! read torque data
    end do

    close ( uid )

!   -----------------------------------------------------------
!   sort in ascending order
!   -----------------------------------------------------------
    gap = ( tqc%ndata + 1) / 2
    do while ( gap >=1 )
        do i = gap+1, tqc%ndata
            do j = i-gap, 1, -gap
!        i = gap
!        do while ( i <= tqc%ndata )
!            j = i - gap
!            do while ( ( j >= 1 ) .and. ( tqc%rev( j ) > tqc%rev( j + gap ) ) )
                if ( tqc%rev( j ) <= tqc%rev( j + gap ) ) exit
                    revt               = tqc%rev(j)
                    tqc%rev( j )       = tqc%rev(j+gap)
                    tqc%rev( j + gap ) = revt
                    tmaxt              = tqc%tq( j )
                    tqc%tq( j )        = tqc%tq( j + gap )
                    tqc%tq( j + gap )  = tmaxt
!                j = j - gap
            end do
!            i = i + 1
        end do
        gap = gap / 2
    end do

    return

!   -----------------------------------------------------------
!   G[Ή
!   -----------------------------------------------------------
600 write (0,'(3A)') " Cannot open torque data file [ ", trim(tqf), " ]"
    stop
650 write (0,'(A,i0,3A)') " Failed to read torque data. Data No.= ",    &
                          tqc%ndata+1, " in [ ", trim( tqf ), " ]"
    close ( uid )
    stop

end subroutine readtqfc



!-----------------------------------------------------------------------
! SUBROUTINE READMAPFC    -- R}bv̓ǂݍ --
!  R}bvf[^̓ǂݍ݃Tu[`
!
!  mapf : R}bvt@C
!  fcm  : R}bv(\)
!-----------------------------------------------------------------------
subroutine readmapfc( uid, mapf, fcm )
    use hvtransfc
    implicit none

    type ( sfcmap ) :: fcm                                      ! R}bvi[
    character ( len=* ) :: mapf                                 ! R}bvt@C
    character ( len=1024 ) :: tmp

    real(8) r, t, f
    integer uid, gap, i, j, k, ios, head, ct

!   -----------------------------------------------------------
!   count number of fuel map data
!   -----------------------------------------------------------
    fcm%ndata = 0                                               ! f[^_̏
    open( uid, file = mapf, status = 'old', err=500 )

    read ( uid, *, err=550 ) tmp                                ! 擪2s̓wb_
    read ( uid, *, err=550 ) tmp
    read ( uid, *, err=550 ) tmp                                ! AChO(XLbv)

    ios = 0
    do while ( ios == 0 )
        read( uid, *, iostat = ios, err=550 ) r, t, f           ! JEgpǂݍ
        if( ios == 0 ) fcm%ndata = fcm%ndata + 1                ! f[^JEg
    end do

    allocate( fcm%rev( fcm%ndata ) )
    allocate( fcm%tq( fcm%ndata )  )
    allocate( fcm%fc( fcm%ndata )  )

!   ----------------
!   - {ԓǂݍ -
!   ----------------
    rewind( uid )                                               ! t@C擪ֈړ
    read ( uid, *, err=550 ) tmp                                ! wb_2s
    read ( uid, *, err=550 ) tmp
!   ----------------------
!   - AChO -
!   ----------------------
    read ( uid, *, err=550 ) r, t, fcm%fcidle                   ! r,t̓_~[lł_

    do i = 1, fcm%ndata                                         ! {ԓǂݍ
        read ( uid, *, err=550 ) fcm%rev(i), fcm%tq(i), fcm%fc(i)
    end do

    close( uid )

!   -----------------------------------------------------------
!   GW]Ƃ\[g
!   -----------------------------------------------------------
    gap = ( fcm%ndata + 1 ) / 2
    do while ( gap >= 1 )
        do i = gap+1, fcm%ndata
            do j = i-gap, 1, -gap
                if ( fcm%rev(j) <= fcm%rev(j+gap) ) exit
                r              = fcm%rev(j)
                fcm%rev(j)     = fcm%rev(j+gap)
                fcm%rev(j+gap) = r
                t              = fcm%tq(j)
                fcm%tq(j)      = fcm%tq(j+gap)
                fcm%tq(j+gap)  = t
                f              = fcm%fc(j)
                fcm%fc(j)      = fcm%fc(j+gap)
                fcm%fc(j+gap)  = f
            end do
        end do
        gap = gap / 2
    end do

!   -----------------------------------------------------------
!   GW]ƂɃgNŏ\[g
!   -----------------------------------------------------------
    head = 1
    r = fcm%rev( head )
    ct = 0
    do i = 1, fcm%ndata
        ct = ct + 1
        if ( r == fcm%rev(i) .and. i < fcm%ndata ) cycle        ! ]̃f[^̓XLbv
!       -------------------------------------------------------
!       ]𒊏oGWgNɕׂȂ
!       -------------------------------------------------------
        if ( r /= fcm%rev(i) ) ct = ct -1                       ! ]ςꍇ̓f[^C
        gap = ( ct + 1 ) / 2
        do while ( gap >= 1 )
            do k = head + gap - 1, head + ct - 1
                do j = k - gap, head, -gap
                    if ( fcm%tq(j) <= fcm%tq(j+gap) ) exit
                    r              = fcm%rev(j)
                    fcm%rev(j)     = fcm%rev(j+gap)
                    fcm%rev(j+gap) = r
                    t              = fcm%tq(j)
                    fcm%tq(j)      = fcm%tq(j+gap)
                    fcm%tq(j+gap)  = t
                    f              = fcm%fc(j)
                    fcm%fc(j)      = fcm%fc(j+gap)
                    fcm%fc(j+gap)  = f
                end do
            end do
            gap = gap / 2
        end do
!       --------------------
!       - ̉]ֈڍs -
!       --------------------
        r = fcm%rev(i)                                          ! ]
        ct = 1                                                  ! f[^
        head = i                                                ! 擪vf
    end do
    return

!   -----------------------------------------------------------
!   error handling
!   -----------------------------------------------------------
500 write (0,'(3A)') " Cannot open FUEL MAP file [ ", trim(mapf), " ]"
    stop
550 write (0,'(3A)') " Failed to read FUEL MAP [ ", trim(mapf), " ]"
    close ( uid )
    stop
end subroutine









!**************************************************************
!  SUBROUTINE OUTAVERAGEFC
!   Output result of average speed & fuel consumption
!**************************************************************
subroutine outaveragefc ( uid, outf, inputf, mapf, ftqf, dist, fuelc, avgv, avgfc )
    use hvtransfc
    implicit none

    character, parameter :: ht = char(9)

    integer uid
    character (len=*) outf, inputf, mapf, ftqf
    real(8) avgv, avgfc, dist, fuelc

    open ( uid, file = outf, status = 'unknown' )

    write ( uid, '(3A)' ) "INPUT_DATA", ht, trim(inputf)
    write ( uid, '(3A)' ) "MAP_NAME", ht, trim(mapf)
    write ( uid, '(3A)' ) "FRQTQ_NAME", ht, trim(ftqf)
    write ( uid, * )
    write ( uid, '(4A,F0.5,3A,F0.4)' ) "TOTAL",ht,"DIST(km)",ht,dist,ht,"FUEL(L)",ht,fuelc
    write ( uid, '(4A,F8.4,3A,F5.1)' ) "AVERAGE",ht,"FC(km/l)",ht,avgfc,ht,"Speed(km/h)",ht,avgv
    write ( uid, * )

    close ( uid )

end subroutine outaveragefc



!**************************************************************
!  SUBROUTINE OUTPUTTIMESERIES
!   Output fuel consumption data of time series
!**************************************************************
subroutine outtimeseriesfc (uid, outf, ndata, t, v, ne, te, sp, fc)
    use hvtransfc
    implicit none

    character, parameter :: ht = char(9)
    integer uid, i, sp(:), ndata
    real(8) t(:), v(:), ne(:), te(:), fc(:)
    character (len=*) outf

    open ( uid, file = outf, status = 'unknown', position = 'APPEND' )

    write ( uid, '(11A)' ) 'time(s)', ht,  'Vreal(km/h)', ht        &
            ,'Ne(rpm)', ht, 'Te(N-m)', ht, 'Shift', ht, 'FC(l/h)'

    do i = 1, ndata
        write (uid, '(f0.2,(a,f0.2),2(a,f0.1),a,i0,a,f0.6)')        &
               t(i), ht,  v(i), ht, ne(i), ht, te(i),               &
               ht, sp(i), ht, fc(i)
    end do

    close ( uid )

end subroutine outtimeseriesfc

